(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

(* The encoding of functions: abstraction and application. *)

(* This encoding is trivial: it has no computational content. *)

(* The meta-theorem that we wish to prove about the encoding is
   as follows. Consider a judgement in the source calculus (say,
   System F with monadic references):

     x1 : T1; x2 : T2; ... |- t : T

   Then, the following judgement holds in the target calculus,
   that is, in FORK:

     w1 : world; x1 : box T1 w1;
     w2 : world; x2 : box T2 (w1 `o` w2);
     ... |- t : box T (w1 `o` w2 `o` ...)

   That is, every lambda-abstraction at the source level introduces
   a new world extension at the target level.

   Let us look at the cases in the proof of this meta-theorem.

   The variable case requires deriving:

     x : box T (w1 `o` ... `o` wk) |- x : box T (w1 `o` ... `o` wn)

   which holds by definition of [box]. (The coercion [forward] proves this.)

   The application case requires deriving:

     ... |- t1 : box (T1 -> T2) w
     ... |- t2 : box T1 w
     --------------------
     ... |- t1 t2 : box T2 w

   The coercion [app], below, proves this. The proof is trivial: just
   remove the outer box by instantiating with the world extension [nil].

   The lambda-abstraction case requires deriving:

     ...; w : world; x : box T1 (... `o` w) |- t : box T2 (... `o` w)
     ----------------------------------------------------------------
     ... |- \x.t : box (T1 -> T2) (...)

   The coercion [abs], below, proves this. The proof is trivial: just
   unfold the definition of [box].

   The unit case is trivial.

   Pair construction and deconstruction are proved by the coercions
   [box_pair] and [pair_box].

   The constants [return], [bind], [new], [read], [write] have been
   encoded in the empty world. Via [forward], they can be moved to
   any world.

   This concludes the review of the cases. *)

(* ------------------------------------------------------------------------- *)

(* Function application. *)

term app:
  forall a b x.
  box (arrow a b) x ->
  box a x ->
  box b x
  =
  /\ a b x.
  \ f : box (arrow a b) x.
  f [nil]

(* ------------------------------------------------------------------------- *)

(* Function abstraction. *)

term abs:
  forall a b x.
  (forall y. box a (x `o` y) -> box b (x `o` y)) ->
  box (arrow a b) x
  =
  /\ a b x.
  \ body : forall y. box a (x `o` y) -> box b (x `o` y).
  body

