Require Import DeBruijn.
Require Import TargetSyntax.

(* ------------------------------------------------------------------------- *)

(* The target language is a pure lambda-calculus. Its semantics is standard. *)

Inductive tred : tterm -> tterm -> Prop :=
| TRedBeta:
    forall t1 t2,
    tred
      (TApp (TAbs t1) t2)
      (subst_tterm t2 0 t1)
| TRedLetPair:
    forall t1 t2 t,
    tred
      (TLetPair (TPair t1 t2) t)
      (subst_tterm t1 0 (subst_tterm (lift_tterm 0 t2) 0 t))
| TRedAbsContext:
    forall t1 t2,
    tred t1 t2 ->
    tred (TAbs t1) (TAbs t2)
| TRedAppLeftContext:
    forall t1 t2 t,
    tred t1 t2 ->
    tred (TApp t1 t) (TApp t2 t)
| TRedAppRightContext:
    forall t1 t2 t,
    tred t1 t2 ->
    tred (TApp t t1) (TApp t t2)
| TRedPairLeftContext:
    forall t1 t2 t,
    tred t1 t2 ->
    tred (TPair t1 t) (TPair t2 t)
| TRedPairRightContext:
    forall t1 t2 t,
    tred t1 t2 ->
    tred (TPair t t1) (TPair t t2)
| TRedLetPairLeftContext:
    forall t1 t2 t,
    tred t1 t2 ->
    tred (TLetPair t1 t) (TLetPair t2 t).

(* ------------------------------------------------------------------------- *)

(* Notions of iterated reduction. *)

Inductive tredstar : tterm -> tterm -> Prop :=
| TRedStarZero:
    forall t,
    tredstar t t
| TRedStarMore:
    forall t1 t2 t3,
    tred t1 t2 ->
    tredstar t2 t3 ->
    tredstar t1 t3.

Inductive tredplus : tterm -> tterm -> Prop :=
| TRedPlus:
    forall t1 t2 t3,
    tred t1 t2 ->
    tredstar t2 t3 ->
    tredplus t1 t3.

(* ------------------------------------------------------------------------- *)

(* Convertibility. *)

Inductive convertible : tterm -> tterm -> Prop :=
| ConvertibleReflexive:
    forall t,
    convertible t t
| ConvertibleRed:
    forall t1 t2,
    tred t1 t2 ->
    convertible t1 t2
| ConvertibleSymmetric:
    forall t1 t2,
    convertible t1 t2 ->
    convertible t2 t1
| ConvertibleTransitive:
    forall t1 t2 t3,
    convertible t1 t2 ->
    convertible t2 t3 ->
    convertible t1 t3.

Local Hint Constructors tred tredstar tredplus convertible.

(* ------------------------------------------------------------------------- *)

(* Basic properties of iterated reduction. *)

Lemma tredstar_transitive:
  forall t1 t2 t3,
  tredstar t1 t2 ->
  tredstar t2 t3 ->
  tredstar t1 t3.
Proof.
  induction 1; eauto.
Qed.

Lemma tredplus_tredstar:
  forall t1 t2,
  tredplus t1 t2 ->
  tredstar t1 t2.
Proof.
  inversion 1; eauto.
Qed.

Hint Resolve tredplus_tredstar.

Lemma tredplus_plus_star:
  forall t1 t2 t3,
  tredplus t1 t2 ->
  tredstar t2 t3 ->
  tredplus t1 t3.
Proof.
  inversion 1. subst. eauto using tredstar_transitive.
Qed.

Lemma tredplus_star_plus:
  forall t1 t2 t3,
  tredstar t1 t2 ->
  tredplus t2 t3 ->
  tredplus t1 t3.
Proof.
  induction 1; inversion 1; subst; eauto using tredstar_transitive.
Qed.

Lemma tredstar_convertible:
  forall t1 t2,
  tredstar t1 t2 ->
  convertible t1 t2.
Proof.
  induction 1; eauto.
Qed.

Lemma inverse_tredstar_convertible:
  forall t1 t2,
  tredstar t2 t1 ->
  convertible t1 t2.
Proof.
  eauto using tredstar_convertible.
Qed.

Lemma tredstar_inverse_tredstar_convertible:
  forall t1 t2 t3,
  tredstar t1 t2 ->
  tredstar t3 t2 ->
  convertible t1 t3.
Proof.
  eauto using tredstar_convertible.
Qed.

(* ------------------------------------------------------------------------- *)

(* Iterated reduction under a context. *)

Lemma TRedStarAbsContext:
  forall t1 t2,
  tredstar t1 t2 ->
  tredstar (TAbs t1) (TAbs t2).
Proof.
  induction 1; eauto.
Qed.

Lemma TRedStarAppLeftContext:
  forall t1 t2 t,
  tredstar t1 t2 ->
  tredstar (TApp t1 t) (TApp t2 t).
Proof.
  induction 1; eauto.
Qed.

Lemma TRedStarAppRightContext:
  forall t1 t2 t,
  tredstar t1 t2 ->
  tredstar (TApp t t1) (TApp t t2).
Proof.
  induction 1; eauto.
Qed.

Lemma TRedStarPairLeftContext:
  forall t1 t2 t,
  tredstar t1 t2 ->
  tredstar (TPair t1 t) (TPair t2 t).
Proof.
  induction 1; eauto.
Qed.

Lemma TRedStarPairRightContext:
  forall t1 t2 t,
  tredstar t1 t2 ->
  tredstar (TPair t t1) (TPair t t2).
Proof.
  induction 1; eauto.
Qed.

Lemma TRedStarLetPairLeftContext:
  forall t1 t2 t,
  tredstar t1 t2 ->
  tredstar (TLetPair t1 t) (TLetPair t2 t).
Proof.
  induction 1; eauto.
Qed.

Lemma TRedPlusAbsContext:
  forall t1 t2,
  tredplus t1 t2 ->
  tredplus (TAbs t1) (TAbs t2).
Proof.
  induction 1; eauto using TRedStarAbsContext.
Qed.

Lemma TRedPlusAppLeftContext:
  forall t1 t2 t,
  tredplus t1 t2 ->
  tredplus (TApp t1 t) (TApp t2 t).
Proof.
  inversion 1; eauto using TRedStarAppLeftContext.
Qed.

Lemma TRedPlusAppRightContext:
  forall t1 t2 t,
  tredplus t1 t2 ->
  tredplus (TApp t t1) (TApp t t2).
Proof.
  induction 1; eauto using TRedStarAppRightContext.
Qed.

Lemma TRedPlusPairLeftContext:
  forall t1 t2 t,
  tredplus t1 t2 ->
  tredplus (TPair t1 t) (TPair t2 t).
Proof.
  induction 1; eauto using TRedStarPairLeftContext.
Qed.

Lemma TRedPlusPairRightContext:
  forall t1 t2 t,
  tredplus t1 t2 ->
  tredplus (TPair t t1) (TPair t t2).
Proof.
  induction 1; eauto using TRedStarPairRightContext.
Qed.

Lemma TRedPlusLetPairLeftContext:
  forall t1 t2 t,
  tredplus t1 t2 ->
  tredplus (TLetPair t1 t) (TLetPair t2 t).
Proof.
  induction 1; eauto using TRedStarLetPairLeftContext.
Qed.

