(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

open Printf
open Atom
open Kinds
open Types

(* A statement is a universally quantified equation between types. *)

type statement =
  | SEquation of locty * locty
  | SForall of atom * kind * statement

type lemmas =
    statement list

(* ------------------------------------------------------------------------- *)

(* Kind checking. *)

(* It is not clear in the syntax at what kind the equality is being
   asserted, so we reject a statement if the two sides have different
   inferred kinds. *)

let rec check env = function
  | SForall (a, kind, s) ->
      check (AtomMap.add a kind env) s
  | SEquation (ty1, ty2) ->
      let kind1 = infer env ty1
      and kind2 = infer env ty2 in
      if not (is_subkind kind1 kind2 && is_subkind kind2 kind1) then
	let loc1, _ = ty1
	and loc2, _ = ty2 in
	Error.signal [ loc1; loc2 ] (sprintf
	  "This equation is ill-kinded, or its kind is ambiguous.\n\
	   The left-hand side has kind:\n\
	   %s\n\
	   The right-hand side has kind:\n\
	   %s"
	  (print_kind kind1)
	  (print_kind kind2)
	)

let check env s =
  try
    check env s
  with KindChecking (locs, msg) ->
    Error.signal locs msg

let check env lemmas =
  List.iter (check env) lemmas;
  Error.signaled()

(* ------------------------------------------------------------------------- *)

(* TEMPORARY nettoyer la-dessous *)

open RawTypes

let play types xenv lemmas =

  let rec p xenv = function
    | SForall (a, _, s) ->
	p (Export.bind xenv a) s
    | SEquation (ty1, ty2) ->
	let ty1 = erase ty1
	and ty2 = erase ty2 in
	(* TEMPORARY si debug, afficher l'enonce avant d'appeler semidecide *)
	match semidecide true types xenv ty1 ty2 with
	| True ->
	    ()
	| Unknown ->
	    printf "Lemma: %s = %s\n%!" (print_ty xenv ty1) (print_ty xenv ty2);
	    printf "     : unknown.\n%!";
	    debug := true;
	    let _ = semidecide false types xenv ty1 ty2 in
	    debug := false
	| False ->
	    printf "Lemma: %s = %s\n%!" (print_ty xenv ty1) (print_ty xenv ty2);
	    printf "     : false!\n%!"
  in

  List.iter (fun s ->
    p xenv s
  ) lemmas

